@@ -4,7 +4,7 @@
 ##
 ##    class "fv" of function value objects
 ##
-##    $Revision: 1.181 $   $Date: 2023/09/08 07:32:16 $
+##    $Revision: 1.185 $   $Date: 2024/04/02 01:23:12 $
 ##
 ##
 ##    An "fv" object represents one or more related functions
@@ -537,6 +537,11 @@ bind.fv <- function(x, y, labl=NULL, desc=NULL, preferred=NULL, clip=FALSE) {
     if(is.null(desc)) desc <- attr(y, "desc")[-yrpos]
     ##
     y <- ystrip
+  } else if(is.function(y)) {
+    ## evaluate the function 'y' at the argument values of the first object
+    xvals <- x[[fvnames(x, ".x")]]
+    yvals <- y(xvals)
+    y <- data.frame(y=yvals)
   } else {
     ## y is a matrix or data frame
     y <- as.data.frame(y)
@@ -590,14 +595,40 @@ cbind.fv <- function(...) {
     ## could be an fv object 
     if(is.fv(a))
       return(a)
+    ## a is a list of arguments
     n <- length(a)
   }
+  ## First argument is template
   z <- a[[1L]]
   if(!is.fv(z))
     stop("First argument should be an object of class fv")
-  if(n > 1)
-    for(i in 2:n) 
-      z <- bind.fv(z, a[[i]])
+  ## Subsequent arguments
+  if(n > 1) {
+    ## save dotnames (only those explicitly given in data -- no 'guessing')
+    dn <- fvnames(z, ".")
+    if(any(isfun <- sapply(a, is.function))) {
+      ## Function(s) will be applied to 'r' values
+      xvals <- z[[fvnames(z, ".x")]]
+      ## Determine variable names using names in cbind call, or default
+      nama <- good.names(names(a), "V", 1:n)
+    }
+    for(i in 2:n) {
+      ai <- a[[i]]
+      if(isfun[i]) {
+        ## apply function to 'r' values
+        yvals <- ai(xvals)
+        ## convert to data frame 
+        ai <- data.frame(y=yvals)
+        colnames(ai) <- nama[i]
+      } else if(is.fv(ai)) {
+        ## collect more dotnames
+        dn <- c(dn, fvnames(ai, "."))
+      }
+      z <- bind.fv(z, ai)
+    }
+    ## reattach dotnames
+    fvnames(z, ".") <- dn
+  }
   return(z)
 }
 
@@ -1181,46 +1212,29 @@ integral.fv <- function(f, domain=NULL, ...) {
   return(ans[1,])
 }
 
-  
-## stieltjes integration for fv objects
+## Stieltjes integration for fv objects
 
-stieltjes <- function(f, M, ...) {
-  ## stieltjes integral of f(x) dM(x)
-  stopifnot(is.function(f))
-  if(is.stepfun(M)) {
-    envM <- environment(M)
-    #' jump locations
-    x <- get("x", envir=envM)
-    #' values of integrand
-    fx <- f(x, ...)
-    #' jump amounts
-    xx <- c(-Inf, (x[-1L] + x[-length(x)])/2, Inf)
-    dM <- diff(M(xx))
-    #' integrate f(x) dM(x)
-    f.dM <- fx * dM
-    result <- sum(f.dM[is.finite(f.dM)])
-    return(list(result))
-  } else if(is.fv(M)) {
-    ## integration variable
-    argu <- attr(M, "argu")
-    x <- M[[argu]]
-    ## values of integrand
-    fx <- f(x, ...)
-    ## estimates of measure
-    valuenames <- names(M) [names(M) != argu]
-    Mother <- as.data.frame(M)[, valuenames]
-    Mother <- as.matrix(Mother, nrow=nrow(M))
-    ## increments of measure
-    dM <- apply(Mother, 2, diff)
-    dM <- rbind(dM, 0)
-    ## integrate f(x) dM(x)
-    f.dM <- fx * dM
-    f.dM[!is.finite(f.dM)] <- 0
-    results <- colSums(f.dM)
-    results <- as.list(results)
-    names(results) <- valuenames
-    return(results)
-  } else stop("M must be an object of class fv or stepfun")
+StieltjesCalc.fv <- function(M, f, ...) {
+  verifyclass(M, "fv")
+  ## integration variable
+  argu <- attr(M, "argu")
+  x <- M[[argu]]
+  ## values of integrand
+  fx <- f(x, ...)
+  ## estimates of measure
+  valuenames <- names(M) [names(M) != argu]
+  Mother <- as.data.frame(M)[, valuenames]
+  Mother <- as.matrix(Mother, nrow=nrow(M))
+  ## increments of measure
+  dM <- apply(Mother, 2, diff)
+  dM <- rbind(dM, 0)
+  ## integrate f(x) dM(x)
+  f.dM <- fx * dM
+  f.dM[!is.finite(f.dM)] <- 0
+  results <- colSums(f.dM)
+  results <- as.list(results)
+  names(results) <- valuenames
+  return(results)
 }
 
 prefixfv <- function(x, tagprefix="", descprefix="", lablprefix=tagprefix,
